/*
 * Program ID:             tree.hpp
 *
 * Description:            Interface definition for in-memory balanced binary
 *                         trees
 *
 * Input Parameters:       none
 *
 * Return/Exit Values:     none
 *
 * Input Files:            none
 *
 * Output Files:           none
 *
 * Linking Procedures:     none
 *
 * Special Logic Notes:    This C++ class uses the AVL (Adelson-Velski and
 *                         Landis) method for correcting binary tree
 *                         imbalances.
 *
 *                         Taken from The C Users Journal, January 1991
 *                         pages 65-74
 *
 *  This class is not able to use a container iterator to delete or detach
 *  Objects stored in the tree.  It results in the corruption of the tree,
 *  and causes a refrence to de-allocated memory.
 *
 **************************************************************************
 *
 *                          Modification Log
 **************************************************************************
 *
 */

/*** revision history ***/
/* 1 TREE.HPP 3-Jul-91,1:42:34,`ERIK' Initial Code Checkin.               */
/* 2 TREE.HPP ,`ERIK' Added virtual to IsA method.                        */
/* 3 TREE.HPP 28-Jul-91,23:01:20,`ERIK' changed some things.              */
/* 4 TREE.HPP 10-Nov-91,21:38:10,`ERIK' Added keyflag field in comment.   */
/* 5 TREE.HPP 10-Nov-91,22:20:16,`ERIK' Integrated with classlib by \     */
/*      inheriting from Object, and making printOn/print_tree pass a Rost */
/* 6 TREE.HPP 10-Nov-91,22:55:08,`ERIK'                                   */
/*      Passed test program, inheriting from \                            */
/*      Object.                                                           */
/* 7 TREE.HPP 10-Nov-91,23:27:18,`ERIK' Removed dsize data element.       */
/* 8 TREE.HPP 10-Nov-91,23:33:12,`ERIK' Got the class working again by \  */
/*      re-adding dsize data element.                                     */
/* 9 TREE.HPP 11-Nov-91,0:27:12,`ERIK' Got everything working again. \    */
/*      However, the comparing of the tree elements does not work.  Shoul */
/*      try inheriting the data pointer from class sortable.              */
/* 10 TREE.HPP 11-Nov-91,12:50:18,`ERIK'                                  */
/*      Passed test program.  Comparing now done \                        */
/*      by calling compare method, which uses compare operators inherited */
/*      Sortable class                                                    */
/* 11 TREE.HPP 11-Nov-91,12:59:36,`ERIK'                                  */
/*      Removed dsize data element, alloc_data, \                         */
/*      copy_data methods.  Passed test program.                          */
/* 12 TREE.HPP 11-Nov-91,13:21:12,`ERIK' Made all the methods return \    */
/*      PCObject rather the Pvoid.  Passed test program.                  */
/* 13 TREE.HPP 11-Nov-91,13:56:36,`ERIK' Make current methods adhere to \ */
/*      Classlib method naming convention.                                */
/* 14 TREE.HPP 11-Nov-91,14:48:06,`ERIK' Spinkled const all over and \    */
/*      made a number of the passed and return types RCObjects to more cl */
/*      match what the classlib is doing.  Ran test program OK.           */
/* 15 TREE.HPP 11-Nov-91,15:22:10,`ERIK' Added elements method.           */
/* 16 TREE.HPP 11-Nov-91,17:18:40,`ERIK' changed keyflag line.            */
/* 17 TREE.HPP 12-Nov-91,0:55:50,`ERIK' Got basic iterator code in place. */
/*      Still need to work out the methods so that they will start, resta */
/*      increment (++), and test (operator int) correctly.                */
/* 18 TREE.HPP 12-Nov-91,11:23:26,`ERIK'                                  */
/*      Got basic iterator functionality.  But \                          */
/*      it does not recognize the end of the tree until it tries to refre */
/*      element after the last, causing an error.  Need to make the int o */
/*      smarter.                                                          */
/* 19 TREE.HPP 12-Nov-91,12:28:46,`ERIK'                                  */
/*      Finally worked out the last problems \                            */
/*      with the iterator.  Passed the test program.                      */
/* 20 TREE.HPP 12-Nov-91,12:50:36,`ERIK'                                  */
/*      Made version control consistent throughout.                       */
/* 21 TREE.HPP 15-Nov-91,15:34:40,`ERIK' removed count.  made findCur     */
/*      And lastError const methods.                                      */
/* 22 TREE.HPP 17-Nov-91,20:50:26,`ERIK' Need to eliminate the \          */
/*      duplication of objects when written to persistent store by        */
/*      the TABSTree class and the TABSNode class.                        */
/* 23 TREE.HPP 10-Dec-91,23:59:12,`ERIK' Removed a blank line.            */
/* 24 TREE.HPP 11-Dec-91,19:07:54,`ERIK' Added error strings and the      */
/*      errorString method to the class.                                  */
/* 25 TREE.HPP 13-Dec-91,0:33:04,`ERIK' moved detach_deletes_data         */
/*      static variable from tree.cpp to class.                           */
/* 26 TREE.HPP 29-Dec-91,10:46:46,`ERIK' ported code to version 3.0       */
/*      of classlib.  Also corrected a number of bugs.                    */
/* 27 TREE.HPP 6-Feb-92,20:50:24,`ERIK' changed appropriate Object        */
/*      arguments and such to Sortable.                                   */
/*** revision history ***/

/***-=>keyflag<=-- "&(#)%n %v %d %t" ***/
/***               "&(#)TREE.HPP 27 6-Feb-92 20:50:24" ***/

/***************************************************************************/

#ifndef __TREE_HPP__
#define __TREE_HPP__

// Standard C headers:
#include <stddef.h>

// Classlib Headers:
#include <contain.h>
#include <sortable.h>

// Project headers:
#include "classnum.hpp"

/***************************************************************************/

typedef enum       TREE_ERR
{  OK,             MEM_ALLOC_FAIL,     NO_DUPES,   TREE_EMPTY,
   NO_SUCH_BalancedTreeNode,   OBJECT_NOT_SORTABLE
}                  TREE_ERR;

/***************************************************************************/

typedef struct     BalancedTreeNode
{
   BalancedTreeNode    *left;
   BalancedTreeNode    *right;
   signed int          balance;
   PSortable           data;       // pointer to object being stored.
}                  BalancedTreeNode;

/***************************************************************************/

_CLASSDEF (BalancedTreeIterator)

_CLASSDEF (BalancedTree)
class _CLASSTYPE BalancedTree : public Container
{
private:
   BalancedTreeNode                *head;
   BalancedTreeNode                *current;
   TREE_ERR            errval;     // error value reported by last operation

   int                 detach_deletes_data;    // module global flag that
                                               // determines if remove will
                                               // delete_data will invoke delete

   void                delete_subtree (BalancedTreeNode *nodeptr);
   void                rebalance (BalancedTreeNode *parent);
   int                 addnode (BalancedTreeNode *parent, BalancedTreeNode *newnode);
   BalancedTreeNode    *findnode (BalancedTreeNode *currnode, RCSortable key);
   int                 remove_node (BalancedTreeNode *currnode, RSortable key);
   int                 getnext (BalancedTreeNode *currnode);
   int                 getprev (BalancedTreeNode *currnode);
   BalancedTreeNode    *BalancedTree::walk_tree (BalancedTreeNode *nptr, int walk_size);

protected:
   virtual void        delete_data (BalancedTreeNode *nodeptr);
   virtual int         compare (RSortable data1, RSortable data2);

   friend class _CLASSTYPE     BalancedTreeIterator;

public:
   BalancedTree  ()
       { head = NULL; current = NULL; errval  = OK; ownsElements (1); }

   ~BalancedTree ()
   {
       detach_deletes_data = 1;
       if (head != NULL)
       {
           delete_subtree (head);
           delete head;
           head = NULL;
       }
   }

   unsigned int    elements () const           { return (getItemsInContainer ()); }
   int             add (RSortable data);
   int             remove (RCSortable data);
   void            detach (RCSortable, int);
   RSortable       find (RCSortable data);
   RCSortable      findFirst (void);
   RCSortable      findNext (void);
   RCSortable      findLast (void);
   RCSortable      findPrev (void);
   int             print_tree (Rostream) const;

   RCSortable      findCurr (void)  const  { return ((RCSortable)*current->data); }
   TREE_ERR        lastError (void) const  { return (errval); }
   const char      *errorString ();


   // Override pure virtual methods, as required by inheriting Object
   virtual classType       isA ()              const { return (BALANCEDTREENUM); }
   virtual char            *nameOf ()          const { return ("BalancedTree");  }
   virtual hashValueType   hashValue ()        const { return (isA ()); }


   virtual void            printOn (Rostream os)  const
                               { print_tree (os); }

   virtual int             isEqual (RCObject O) const
   {
       int retval = 0;
       if (this == &O)
           retval = 1;
       return (retval);
   }

   // Additional methods to make this class look more like a Classlib
   // container class.                 
   virtual int             hasMember (RCSortable O)
                               {
                                   if (find (O) == NOOBJECT)
                                       return (0);
                                   else
                                       return (1);
                               }

   virtual RSortable         findMember (RCSortable O)
                               { return (find (O)); }

   void                    destroy (RCSortable o) { remove (o); }

   virtual void            flush (DeleteType = DefDelete) { }

   virtual RContainerIterator initIterator () const;
};

/***************************************************************************/

class _CLASSTYPE BalancedTreeIterator : public ContainerIterator
{
public:
   BalancedTreeIterator (RCBalancedTree);

   virtual ~BalancedTreeIterator ()
   { }

   virtual operator            int ();
   virtual RObject             current () { return ((RObject)beingIterated.findCurr ()); }
   virtual operator            RObject ();
   virtual RObject operator    ++ ();
   virtual RObject operator    ++ (int);
   virtual void                restart ();

private:
   RCBalancedTree  beingIterated;
   int             firstflag;
   PCObject        last;
};

#endif

